@@ -130,10 +130,13 @@ class YandexRecognizer(YandexBaseBackend):
         timer = threading.Timer(300, listener.stop)
         timer.start()
         self.partials_counter = 0
+        text_buffer = None
         for answer in self._recognize(listener):
             # Каждый раз получаем самый последний буфер, т.к. он может быть
-            # создан в процессе работы.
-            text_buffer = get_text_buffer()
+            # создан в процессе работы. Но для сохранения финальных результатов
+            # используем запомненный.
+            if text_buffer is None:
+                text_buffer = get_text_buffer()
             chunk = answer.chunks[0]
             alternatives = chunk.alternatives
             text = alternatives[0].text
@@ -145,6 +148,7 @@ class YandexRecognizer(YandexBaseBackend):
                 )
                 text_to_text_buffer(text, fixed=True, text_buffer=text_buffer)
                 self.partials_counter = 0
+                # После каждого финального текста сбрасываем привязку к буферу.
                 text_buffer = None
             else:
                 logger_debug(
