@@ -18,7 +18,7 @@ from tempfile import NamedTemporaryFile
 from time import time, sleep
 from threading import Thread
 
-from astersay.exceptions import AgiError, AgiSigpipeError
+from astersay.exceptions import AgiError, AgiHangupError, AgiSigpipeError
 
 logger = logging.getLogger(__name__)
 log_debug = logger.debug
@@ -57,9 +57,9 @@ def _make_say_params(saydata, variables):
     params = {'text': '', 'escape_digits': '', 'voice': ''}
 
     if 'voice' in saydata:
-        params['voice'] = saydata['voice']
+        params['voice'] = saydata['voice'] % variables
     elif 'file' in saydata:
-        params['voice'] = splitext(saydata['file'])[0]
+        params['voice'] = splitext(saydata['file'])[0] % variables
     else:
         params['text'] = saydata['text'] % variables
 
@@ -87,6 +87,47 @@ def _make_pause(data, key):
     return pause
 
 
+def _prepare_answer(answer, export_name, script, variables, source=None):
+    export_name_source = export_name + '_source'
+    if source is True:
+        variables[export_name_source] = answer
+    elif source is not None:
+        variables[export_name_source] = source
+
+    booleanize = script.get('booleanize')
+    convert = script.get('convert')
+    if booleanize:
+        if booleanize.get('to_lowercase', True):
+            answer = answer.lower()
+        answer = bool(answer in booleanize['true_list'])
+        text_true = booleanize.get('text_true') or _('Да')
+        text_false = booleanize.get('text_false') or _('Нет')
+        export_name_text = (
+            booleanize.get('export_name_text') or
+            export_name + '_text'
+        )
+        variables[export_name_text] = text_true if answer else text_false
+    elif convert:
+        variables[export_name + '_convert_from'] = answer
+        if convert.get('to_lowercase', True):
+            answer = answer.lower()
+        default = convert.get('default')
+        if default is None:
+            default = answer
+        answer = convert['values'].get(answer, default)
+        labels = convert.get('labels')
+        if labels and answer in labels:
+            variables[export_name + '_label'] = labels[answer]
+        names = convert.get('names')
+        if names and answer in names:
+            variables[export_name + '_name'] = labels[answer]
+        texts = convert.get('texts')
+        if texts and answer in texts:
+            variables[export_name + '_text'] = labels[answer]
+    # Устанавливаем изменённый текст ответа в главную переменную.
+    variables[export_name] = answer
+
+
 class Dialog:
     """Диалог робота с позвонившим на номер абонентом."""
     is_closed = False
@@ -112,6 +153,8 @@ class Dialog:
         self.text_buffers = OrderedDict()
         # Словарь для экспортируемых после завершения в Asterisk переменных.
         self.export = OrderedDict()
+        # Цепочка исполнения скриптов с их результатами и распознанным текстом.
+        self.chain = []
 
     def make_text_buffer(self, key=None):
         if not key:
@@ -168,6 +211,14 @@ class Dialog:
         text = self.text_from_text_buffer(mode, text_buffer)
         return regexp.search(text), text
 
+    def capture_text(self, stopwords, mode=1, text_buffer=None):
+        """Метод ищет в тексте буфера по регулярному выражению."""
+        text = self.text_from_text_buffer(mode, text_buffer)
+        stop = False
+        if stopwords and stopwords.intersection(text.split(' ')):
+            stop = True
+        return text, stop
+
     def text_processing(self, name, mode=1, text_buffer=None):
         """
         Метод отправляет текст буфера в подпроцесс обработчика текста и
@@ -178,6 +229,13 @@ class Dialog:
             return '', text
         text = text.replace('\n', ' ')
         processor = self.settings.get_processor(name)
+        if getattr(self, '_text_processing_processor', '') != processor:
+            self._text_processing_processor = processor
+            self._text_processing_result = self._text_processing_text = ''
+        # Кэшированое значение результата для текста.
+        if self._text_processing_text == text:
+            return self._text_processing_result, self._text_processing_text
+
         procname = basename(processor)
         log_debug('> PROCESSOR %s %r', procname, text)
         res = subprocess.run(
@@ -193,6 +251,8 @@ class Dialog:
         else:
             result = res.stdout.decode('utf-8').rstrip()
             log_debug('< PROCESSOR %s %r', procname, result)
+        self._text_processing_result = result
+        self._text_processing_text = text
         return result, text
 
     def say(self, text: str, voice: str, escape_digits: str, nonblocking=False):
@@ -312,12 +372,19 @@ class Dialog:
         self.make_text_buffer()
         # Затем запускаем процесс потокового распознавания речи.
         self.start()
+        hangup = False
+        error = None
         try:
             self.main_script()
+        except AgiHangupError:
+            log_info(_('Абонент повесил трубку до окончания диалога.'))
+            hangup = True
         except AgiError as e:
-            log_error(_('Ошибка в AGI.'), exc_info=e)
+            error = _('Ошибка в AGI.')
+            log_error(error, exc_info=e)
         except Exception as e:
-            log_error(_('Ошибка в сценарии диалога.'), exc_info=e)
+            error = _('Ошибка в сценарии диалога.')
+            log_error(error, exc_info=e)
         self.stop()
 
         agi = self.agi
@@ -329,7 +396,10 @@ class Dialog:
                 _('Экспорт %(key)s=%(value)s'), {'key': key, 'value': value})
             agi.set(key, value)
         filename = self.settings.export_dialog(
-            self.agi.params, self.export, self.text_buffers
+            self.agi.params, self.export, self.text_buffers,
+            hangup=hangup,
+            error=error,
+            chain=self.chain,
         )
         if filename:
             log_info(_('Диалог экспортирован в файл %s'), filename)
@@ -341,11 +411,55 @@ class Dialog:
 
         log_info(_('Диалог завершён.'))
 
+    def execute_plugin(self, name, result, speech, variables):
+        """
+        Возвращает кортеж из результата выполнения подпроцесса и словаря данных,
+        которыми плагин оперировал.
+        Плагин может устанавливать свои переменные в процессе своей работы или
+        ещё какие-либо данные в файл обмена данными.
+        """
+        plugin = self.settings.get_plugin(name)
+        log_debug('> PLUGIN %s', plugin)
+        plugin_data = {
+            'agi_params': self.agi.params,
+            'work_dir': self.settings.work_dir,
+            'speech': speech,
+            'variables': variables,
+            'chain': self.chain,
+            'text_buffers': self.text_buffers,
+        }
+        if isinstance(result, (str, list, tuple, int, float)):
+            plugin_data['result'] = result
+        else:
+            plugin_data['result'] = str(result)
+        tempfile = NamedTemporaryFile(mode='w', suffix='.json')
+        tempfile_name = tempfile.name
+        json.dump(plugin_data, tempfile)
+        tempfile.flush()
+        log_debug('JSONFILE %s', tempfile_name)
+        res = subprocess.run(
+            [plugin, tempfile_name],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            universal_newlines=True,
+        )
+
+        if res.returncode:
+            log_error(_('Плагин %r сломан.'), plugin)
+            log_error('< PLUGIN %s', res.stderr)
+        else:
+            log_debug('< PLUGIN %s', res.stdout)
+            plugin_data = json.load(open(tempfile_name))
+            variables.update(plugin_data['variables'])
+        tempfile.close()
+        return res, plugin_data
+
     def main_script(self):
         started = time()
         scheme = self.settings.dialog_scheme
         constants = scheme.get('constants', {})
         self.export = variables = OrderedDict()
+        self.chain = chain = []
         variables.update(constants)
         # Словарь скриптов.
         scripts = scheme.get('scripts', {})
@@ -363,52 +477,31 @@ class Dialog:
             # Здесь должна возникать ошибка AttributeError для сломанных схем.
             method = getattr(self, '_script_%s' % scrypt_type)
             log_debug('  SCRIPT %s', name.upper())
+            self.partials_begin = 0
             if 'partials_begin' in script:
                 try:
                     self.partials_begin = int(script['partials_begin'])
                 except ValueError:
-                    self.partials_begin = 0
+                    pass
+
+            say_debug = process.get('say_debug')
+            if say_debug:
+                saydata = {'text': _('Запускаю скрипт %(name)r') % process}
+                self.say(**_make_say_params(saydata, variables))
+
             result, speech = method(script, variables)
             log_debug('  RESULT %r', result)
             log_debug('  SPEECH %r', speech)
+            # Экспортируем разговор в отдельную переменную, чтобы иметь
+            # доступ к нему в других скриптах и плагинах.
+            export_speech = script.get('export_speech')
+            if export_speech:
+                variables[export_speech] = speech
 
             # Запускаем плагин при его наличии.
             plugin_name = process.get('plugin')
             if plugin_name:
-                plugin = self.settings.get_plugin(plugin_name)
-                log_debug('> PLUGIN %s', plugin)
-                plugin_data = {
-                    'agi_params': self.agi.params,
-                    'work_dir': self.settings.work_dir,
-                    'speech': speech,
-                    'variables': variables,
-                    'text_buffers': self.text_buffers,
-                }
-                if isinstance(result, (str, list, tuple, int, float)):
-                    plugin_data['result'] = result
-                else:
-                    plugin_data['result'] = str(result)
-                tempfile = NamedTemporaryFile(mode='w', suffix='.json')
-                tempfile_name = tempfile.name
-                json.dump(plugin_data, tempfile)
-                tempfile.flush()
-                log_debug('JSONFILE %s', tempfile_name)
-                res = subprocess.run(
-                    [plugin, tempfile_name],
-                    stdout=subprocess.PIPE,
-                    stderr=subprocess.PIPE,
-                    universal_newlines=True,
-                )
-                if res.returncode:
-                    log_error(_('Плагин %r сломан.'), plugin)
-                    log_error('< PLUGIN %s', res.stderr)
-                else:
-                    log_debug('< PLUGIN %s', res.stdout)
-                    # Плагин может установить свои переменные в процессе своей
-                    # работы.
-                    plugin_data = json.load(open(tempfile_name))
-                    variables.update(plugin_data['variables'])
-                tempfile.close()
+                self.execute_plugin(plugin_name, result, speech, variables)
 
             # Устанавливаем следующий процесс.
             success = process.get('success')
@@ -419,14 +512,36 @@ class Dialog:
             if success and '%(' in success:
                 success %= variables
             fail = process.get('fail')
+            if fail and '%(' in fail:
+                fail %= variables
             log_debug(' SUCCESS %s', success)
             log_debug('    FAIL %s', fail)
             if process.get('finish', False):
                 log_debug('  FINISH')
+                if say_debug:
+                    saydata = {'text': _('Завершаю диалог.')}
+                    self.say(**_make_say_params(saydata, variables))
                 process = None
             elif result and success in program:
+                if say_debug:
+                    saydata = {'text': _('Исходная речь. %s.') % speech}
+                    self.say(**_make_say_params(saydata, variables))
+                    if result is True:
+                        saydata = {'text': _('В результате получена истина.')}
+                    else:
+                        saydata = {
+                            'text': _('Разультат обработки. %s.') % result}
+                    self.say(**_make_say_params(saydata, variables))
                 process = program[success]
             elif not result and fail in program:
+                if say_debug:
+                    if speech:
+                        saydata = {'text': _('Исходная речь. %s.') % speech}
+                    else:
+                        saydata = {'text': _('Исходная речь пуста.')}
+                    self.say(**_make_say_params(saydata, variables))
+                    saydata = {'text': _('Результат отсутствует.')}
+                    self.say(**_make_say_params(saydata, variables))
                 process = program[fail]
             else:
                 try:
@@ -434,9 +549,14 @@ class Dialog:
                 except IndexError:
                     process = None
             log_debug('    NEXT %s', process)
+            if say_debug and process:
+                saydata = {'text': _('Следующий скрипт %(name)r') % process}
+                self.say(**_make_say_params(saydata, variables))
 
             # Восстанавливаем константы. Это проще, чем городить лес кода.
             variables.update(constants)
+            # Сохраняем цепочку вызовов скриптов.
+            chain.append([name, result, speech])
 
         finished = time()
         variables['_time_start_'] = started
@@ -493,6 +613,123 @@ class Dialog:
         self._voice_duration = 0
         log_debug(_('Прерывание воспроизведения выполнено.'))
 
+    def _script_capture(self, script, variables):
+        assert script['type'] == 'capture'
+
+        assert 'export_name' in script
+        export_name = script['export_name']
+        assert export_name and isinstance(export_name, str)
+
+        pauses = script.get('pauses', {})
+        _make_pause(pauses, 'before')
+
+        commands = script.get('commands', {})
+        self._agi_execute(commands, 'before')
+
+        mode = int(script.get('mode_buffer', 1))
+        assert mode in (1, 2, 3)
+        # Для капчуринга всегда создаём новый буфер.
+        self.make_text_buffer()
+
+        saydata = script.get('say_start')
+        if saydata:
+            _make_pause(pauses, 'before_say_start')
+            self._agi_execute(commands, 'before_say_start')
+            self.say(**_make_say_params(saydata, variables))
+            _make_pause(pauses, 'after_say_start')
+            self._agi_execute(commands, 'after_say_start')
+
+        maxtime = float(script.get('maxtime', 600))
+        minlength = int(script.get('minlength', 10))
+        silence = int(script.get('silence', 15))
+        stopwords = script.get('stopwords', [])
+        log_debug(_('Ограничение по времени: %s секунд.'), maxtime)
+        log_debug(_('Ограничение по длине: от %s символов.'), minlength)
+        log_debug(_('Ограничение по тишине: %s секунд.'), stopwords)
+        log_debug(_('Ограничение по стоп-словам: %s.'), stopwords)
+        stopwords = set(stopwords)
+
+        func = self.capture_text
+        speech, stop = func(stopwords, mode=mode)
+        # Переменная для определения молчания.
+        _speech = speech
+        if stop:
+            log_debug(_('Первая попытка успешна. Текст: %s'), speech)
+        else:
+            log_debug(_('Первая попытка провалилась. Текст: %s'), speech)
+        # В блокирующем режиме голоса его длина уже будет сброшена в 0.
+        maxtime += self._voice_duration
+
+        second = 0
+        # Количество секунд тишины.
+        silence_second = 0
+        while not stop and second < maxtime:
+            log_debug(_('Стоп-слова пока нет, ждём %d-ю секунду.'), second)
+            second += 0.5
+            sleep(0.5)
+            silence_second += 0.5
+            speech, stop = func(stopwords, mode=mode)
+            if stop:
+                log_debug('Сработало ограничение по стоп-слову.')
+            else:
+                _expire = silence_second >= silence
+                _size = len(speech) >= minlength
+                stop = _expire and _size
+                if _expire:
+                    log_debug('Сработало ограничение по тишине.')
+                    if _size:
+                        log_debug('Сработало ограничение по длине.')
+                    else:
+                        log_debug('Длина текста недостаточна для завершения.')
+            if _speech != speech:
+                # log_debug('old speech %r', _speech)
+                # log_debug('new speech %r', speech)
+                silence_second = 0
+                _speech = speech
+            else:
+                log_debug(_('Тишина %r секунд.'), silence_second)
+                pass
+            if self.is_closed:
+                log_debug(_('Диалог завершился до приёма данных.'))
+                break
+
+        if speech:
+            answer = speech
+            log_debug(_('Ответ: %r'), answer)
+            _prepare_answer(
+                answer=answer, source=answer, export_name=export_name,
+                script=script, variables=variables)
+
+            self._agi_stop_playback(script.get('say_start'))
+
+            # Когда указан файл плагина, то он должен обработать переменные
+            # перед say_success.
+            plugin_name = script.get('plugin')
+            if plugin_name:
+                self.execute_plugin(plugin_name, answer, speech, variables)
+
+            saydata = script.get('say_success')
+            if saydata:
+                _make_pause(pauses, 'before_say_success')
+                self._agi_execute(commands, 'before_say_success')
+                self.say(**_make_say_params(saydata, variables))
+                _make_pause(pauses, 'after_say_success')
+                self._agi_execute(commands, 'after_say_success')
+        else:
+            answer = None
+            log_debug(_('Ответа нет.'))
+            saydata = script.get('say_fail')
+            if saydata:
+                _make_pause(pauses, 'before_say_fail')
+                self._agi_execute(commands, 'before_say_fail')
+                self.say(**_make_say_params(saydata, variables))
+                _make_pause(pauses, 'after_say_fail')
+                self._agi_execute(commands, 'after_say_fail')
+
+        _make_pause(pauses, 'after')
+        self._agi_execute(commands, 'after')
+        return answer, speech
+
     def _script_parse_names(self, script, variables):
         assert script['type'] == 'parse_names'
 
@@ -548,7 +785,9 @@ class Dialog:
                 length_names = int(script.get('length_names', 2))
                 answer = ' '.join(names[:length_names])
                 log_debug(_('Ответ: %r'), answer)
-                variables[export_name] = answer
+                _prepare_answer(
+                    answer=answer, source=answer, export_name=export_name,
+                    script=script, variables=variables)
 
                 self._agi_stop_playback(script.get('say_start'))
 
@@ -647,32 +886,9 @@ class Dialog:
             if match:
                 answer = match.group()
                 log_debug(_('Ответ: %r'), answer)
-                booleanize = script.get('booleanize')
-                convert = script.get('convert')
-                if booleanize:
-                    if booleanize.get('to_lowercase', True):
-                        answer = answer.lower()
-                    answer = bool(answer in booleanize['true_list'])
-                    text_true = booleanize.get('text_true') or _('Да')
-                    text_false = booleanize.get('text_false') or _('Нет')
-                    export_name_text = (
-                        booleanize.get('export_name_text') or
-                        export_name + '_text'
-                    )
-                    variables[export_name_text] = text_true if answer else text_false
-                elif convert:
-                    variables[export_name + '_source'] = answer
-                    if convert.get('to_lowercase', True):
-                        answer = answer.lower()
-                    default = convert.get('default')
-                    if default is None:
-                        default = answer
-                    answer = convert['values'].get(answer, default)
-                    labels = convert.get('labels')
-                    if labels and answer in labels:
-                        variables[export_name + '_label'] = labels[answer]
-
-                variables[export_name] = answer
+                _prepare_answer(
+                    answer=answer, source=answer, export_name=export_name,
+                    script=script, variables=variables)
 
                 self._agi_stop_playback(script.get('say_start'))
 
@@ -706,6 +922,10 @@ class Dialog:
 
         _make_pause(pauses, 'after')
         self._agi_execute(commands, 'after')
+        if script.get('fail_on_false'):
+            answer = variables.get(export_name) in (
+                True, 'True', 'true', 'TRUE')
+            log_debug(_('Режим `fail_on_false`, ответ %r.'), answer)
         return answer, speech
 
     def _script_text_processing(self, script, variables):
@@ -766,8 +986,9 @@ class Dialog:
             if response:
                 answer = response
                 log_debug(_('Ответ: %r'), answer)
-                variables[export_name] = answer
-                variables[export_name + '_source'] = speech
+                _prepare_answer(
+                    answer=answer, source=speech, export_name=export_name,
+                    script=script, variables=variables)
 
                 self._agi_stop_playback(script.get('say_start'))
 
